"""CJK word splitter for use with ZCTextIndex in Zope.

DESCRIPTION OF ALGORITHM

  Continuous letters and numbers make up words.  Spaces and symbols
  separate letters and numbers into words.  This is sufficient for
  all western text.

  CJK doesn't use spaces or separators to separate words, so the only
  way to really find out what constitutes a word would be to have a
  dictionary and advanced heuristics.  Instead, we form pairs from
  consecutive characters, in such a way that searches will find only
  characters that appear more-or-less the right sequence.  For example:

    ABCDE => AB BC CD DE

  This works okay since both the index and the search query is split
  in the same manner, and since the set of characters is huge so the
  extra matches are not significant.


TODO
  - isCJK may be too wide; we want Hanzi/Katakana only, which does
    not use any separators characters between words.
  - Improve performance.
  - Support globbing; doesn't really make sense since each word is only
    two characters long.
"""

from Products.ZCTextIndex.ISplitter import ISplitter
from Products.ZCTextIndex.PipelineFactory import element_factory
from unicodedata import category

# this should probably be configurable through the web
# it is, in general, a bad idea to rely on this kind of automatic
# conversion; instead, try to always pass real unicode strings.
default_encoding = "ASCII"      # encoding to assume non-Unicode text to have
nonmappable_handling = "strict" # "strict", "ignore", or "replace";
                                # see unicode() function documentation


def isCJK(ch):
    """...is CJK character?"""
    return category(ch) in ["Lo"]           # "Letter, Other" seems to be CJK stuff

def isletnum(ch):
    """...is letter or number?"""
    return category(ch)[0] in ['L', 'N']



class CJKSplitter:

    __implements__ = ISplitter
    
    def process(self, text, isGlob=0):
        isCJKWord = 0
	words = []
        word = u""

        if type(text) in [type([]), type(())]:
            text = " ".join(text)

        if type(text) == type(""):
            # convert to Unicode string
            text = unicode(text, default_encoding, nonmappable_handling)

        for ch in text:
            if not isletnum(ch):
                words += [word]
                word = u""
                continue

            if isCJKWord:
                if isCJK(ch):
                    word += ch
                    words += [word]
                    word = ch
                else:
                    words += [word]
                    word = ch
                    isCJKWord = 0
            else:
                if isCJK(ch):
                    words += [word]
                    word = ch
                    isCJKWord = 1
                else:
                    word += ch

        if word: # get the left-over last word as well
            words += [word]

        return filter(lambda x: len(x) > 1, words)

#    def processGlob(self, text):       not supported yet
#        return self.process(text, 1)

try:
  element_factory.registerFactory(
      "Word Splitter", "CJK splitter", CJKSplitter)
except ValueError:
    pass # this is probably caused by a refresh
